import gzip
from pathlib import Path
from typing import Callable

from helix.configuration import MANAGER_CFG
from helix.data.microarray_converter import MicroarrayConverterTarget
from helix.microarray.raw_file import RawEntry, RawFile
from helix.naming.converter import Converter


class MicroarrayConverter:
    def __init__(self, config=MANAGER_CFG.REPOSITORY) -> None:
        self._config = config
        self._template_folder = self._config.metadata.joinpath("microarray_templates")
        self._comments = []

        if not self._template_folder.exists():
            raise FileNotFoundError(
                f"Unable to find microarray template body folder at {self._template_folder!s}"
            )

    def _null_result(self, target: MicroarrayConverterTarget):
        if target in [
            MicroarrayConverterTarget.Ancestry_v1,
            MicroarrayConverterTarget.Ancestry_v2,
        ]:
            return "00"
        return "--"

    def ingest(self, input: Path, progress: Callable[[str, float], None] = None):
        name_noext = input.name.removesuffix("".join(input.suffixes))
        target = self._template_folder.joinpath(name_noext).with_suffix(".txt.gz")

        if target == input:
            raise ValueError(f"Input file is the same as the target: {input!s}")

        if not input.exists():
            raise FileNotFoundError(
                f"Unable to find body file for microarray template at {input!s}"
            )

        rawfile = RawFile(input)
        template = rawfile.load()

        entries = []
        with gzip.open(target, "wt") as file:
            if template.meta is not None:
                file.write(f"##{template.meta.model_dump_json()}\n")
            file.writelines(template.comments)
            for chromosome in Converter.sort(list(template.grouped_entries.keys())):
                ordered_positions = list(template.grouped_entries[chromosome].keys())
                ordered_positions.sort()
                for position in ordered_positions:
                    entries = template.grouped_entries[chromosome][position]
                    for entry in entries:
                        file.write(template.meta.input_format.format(**entry.__dict__))
        return target

    def convert(self, input: Path, target: Path):
        if not target.exists():
            raise FileNotFoundError(
                f"Unable to find body file for microarray template at {target!s}"
            )

        template = RawFile(target).load()
        query_entries = RawFile(input).load()

        lines = []
        for chromosome in template.grouped_entries:
            for position in template.grouped_entries[chromosome]:
                template_entries: set[RawEntry] = template.grouped_entries[chromosome][
                    position
                ]
                query_entry = None
                if chromosome not in query_entries.grouped_entries:
                    query_entry = template.meta.undetermined
                elif position not in query_entries.grouped_entries[chromosome]:
                    query_entry = template.meta.undetermined
                else:
                    if len(query_entries.grouped_entries[chromosome][position]) == 0:
                        raise RuntimeError(
                            f"Bug: could not find a query entry for {chromosome} {position}"
                        )
                    entries = query_entries.grouped_entries[chromosome][position]
                    entries = list(entries)
                    query_entry = entries[0].result

                for template_entry in template_entries:
                    output_line = template.meta.output_format.format(
                        id=template_entry.id,
                        chromosome=template_entry.chromosome,
                        position=template_entry.position,
                        result=query_entry,
                    )
                    lines.append(output_line)

        suffix = template.meta.file_extension
        stem = input.stem.replace("_CombinedKit", "")
        output = input.with_name(f"{stem}_{target.name}{suffix}")

        with output.open("a", newline="\n") as f:
            f.writelines(template.comments)
            f.writelines(lines)
        return output



import shutil

import pytest
from helix.data.genome import Genome
from helix.files.bgzip import BGzip
from helix.files.file_type_checker import FileType, FileTypeChecker
from test.genome_fixtures import format_map, remote_repo_fixture


def test_compress_gzip(remote_repo_fixture, tmp_path_factory):
    temp_folder = tmp_path_factory.mktemp("tmp") / format_map["gzip"]
    genome: Genome = remote_repo_fixture["gzip"]["genome"]
    local_repo_folder = tmp_path_factory.mktemp("genomes")
    genome.parent_folder = local_repo_folder
    shutil.copy(remote_repo_fixture["gzip"]["file_on_disk"], temp_folder)
    sut = BGzip()
    with pytest.raises(RuntimeError) as exc:
        output = sut.perform(genome, temp_folder)
    assert "not decompressed" in str(exc.value)


def test_compress_fasta(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["fasta"]
    genome: Genome = remote_repo_fixture["fasta"]["genome"]
    local_repo_folder = tmp_path_factory.mktemp("genomes")
    genome.parent_folder = local_repo_folder
    shutil.copy(remote_repo_fixture["fasta"]["file_on_disk"], target)
    sut = BGzip()
    output = sut.perform(genome, target)
    assert FileTypeChecker().get_type(output) == FileType.BGZIP
    assert genome.gzi.exists()


def test_compress_bgzip(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["bgzip"]
    genome: Genome = remote_repo_fixture["bgzip"]["genome"]
    local_repo_folder = tmp_path_factory.mktemp("genomes")
    genome.parent_folder = local_repo_folder
    shutil.copy(remote_repo_fixture["bgzip"]["file_on_disk"], target)
    sut = BGzip()
    output = sut.perform(genome, target)
    assert FileTypeChecker().get_type(output) == FileType.BGZIP
    assert genome.gzi.exists()


def test_compress_zip(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["zip"]
    genome: Genome = remote_repo_fixture["zip"]["genome"]
    local_repo_folder = tmp_path_factory.mktemp("genomes")
    genome.parent_folder = local_repo_folder
    shutil.copy(remote_repo_fixture["zip"]["file_on_disk"], target)
    sut = BGzip()
    with pytest.raises(RuntimeError) as exc:
        sut.perform(genome, target)
    assert "not decompressed" in str(exc.value)


def test_compress_razf(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["razf"]
    genome: Genome = remote_repo_fixture["razf"]["genome"]
    local_repo_folder = tmp_path_factory.mktemp("genomes")
    genome.parent_folder = local_repo_folder
    shutil.copy(remote_repo_fixture["razf"]["file_on_disk"], target)
    sut = BGzip()
    with pytest.raises(RuntimeError) as exc:
        sut.perform(genome, target)
    assert "not decompressed" in str(exc.value)



import subprocess
import sys
from xml.etree import ElementTree

import requests
from packaging.version import Version

from helix import VERSION


class Updater:

    def __init__(self) -> None:
        self.packages = ["doublehelix", "doublehelix-external"]

    def check(self, stable_only=False):
        r = requests.get("https://pypi.org/rss/project/doublehelix/releases.xml")
        r.raise_for_status()
        rss_feed = ElementTree.fromstring(r.text)
        channel = rss_feed.find("channel")
        versions = []
        for item in channel.findall("item"):
            title = item.find("title").text
            version = Version(title)
            versions.append(version)

        if stable_only:
            filtered_versions = []
            for version in versions:
                v = Version(version)
                if v.is_devrelease or v.is_postrelease or v.is_prerelease:
                    continue
                filtered_versions.append(version)
            versions = filtered_versions

        if len(versions) == 0:
            return None

        versions.sort()

        latest = versions[-1]
        current = Version(VERSION.version)

        # In case you're using a dev version of a public release, that's the most
        # recent one.
        if latest.pre == current.pre and latest.base_version == current.base_version:
            return None

        if latest > current:
            return latest
        return None

    def update(self, package_name, versions):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                f"{package_name}=={versions}",
            ]
        )
        print(f"Upgraded {package_name} from {versions[0]} to {versions[1]}")

        sys.exit()


if __name__ == "__main__":
    updater = Updater()
    latest = updater.check()
    if latest is not None:
        updater.update(latest)



from typing import Callable, List
from helix.fasta.letter_run import LetterRun


class LetterRunCollection:
    """Represent a collection of letter's run in a specific sequence

    Example:
        The following example indicates the repetition of 10 identical letter
        in the chromosome "chr1" that starts at position 0 and ends in the
        position 10:
        >>> collection = LetterRunCollection(name="chr1", length=10000)
        >>> collection.open_run(0)
        >>> collection.close_run(10)
        >>> collection.end(10000)

        The exact letter the run indicates is not tracked inside this class.
    """

    def __init__(self, name: str, length: int) -> None:
        """Initialize the collection

        Args:
            name (str): Name of the sequence
            length (int): Length of the sequence

        Raises:
            IndexError: Length is less than or equal to zero
        """
        if length <= 0:
            raise IndexError("Length should be greater than zero.")
        self.runs: List[LetterRun] = []
        self.name: str = name
        self.length: int = length
        self._current_run: LetterRun = None
        self._is_ended: bool = False

    def filter(self, criteria: Callable[[LetterRun], bool]) -> List[LetterRun]:
        """Filter the runs

        Args:
            criteria (typing.Callable[[LetterRun], bool]): Criteria to filter the runs

        Returns:
            _type_: Filtered runs
        """
        return [x for x in self.runs if criteria(x)]

    def open_run(self, position: int) -> None:
        """Indicates the beginning of a new run.

        Args:
            position (int): Position of the beginning of the new run

        Raises:
            RuntimeError: open_run was called after end() is called, meaning
                the collection is closed because the sequence is finished.
            ValueError: open_run was called on a run that was already opened.
            RuntimeError: open_run was called when the caller did not call
                close_run() for the previous run.
            ValueError: If position is before the end of the previous run
        """
        if self._is_ended:
            raise RuntimeError("Trying to open a run after the sequence was closed")
        if self._current_run is not None:
            raise RuntimeError("Trying to open an already open run")
        if position < 0:
            raise ValueError("Trying to open a run with a negative position")

        if len(self.runs) > 0:
            previous_end = self.runs[-1].start + self.runs[-1].length
            if position < previous_end + 1:
                raise ValueError(
                    "Position cannot be before the end of the previous run"
                )

        self._current_run = LetterRun()
        self._current_run.open(position)

    def is_run_open(self) -> bool:
        """Checks whether a run was opened.

        Returns:
            bool: True if the run is open.
        """
        return self._current_run is not None

    def close_run(self, position: int) -> None:
        """Indicates the end of a run.

        Args:
            position (int): Position where the run ends.

        Raises:
            RuntimeError: close_run called before calling open_run first
            RuntimeError: Position cannot be less then the start of this run
            RuntimeError: Position is greater than the end of the sequence
        """
        if self._current_run is None:
            raise RuntimeError("Trying to close an already closed run.")
        if position <= self._current_run.start:
            raise ValueError(
                f"Position cannot be smaller than start: {position}<={self._current_run.start}"
            )
        if position > self.length:
            raise ValueError(
                "Position cannot be greater than the length of the sequence: "
                f"{position}>{self.length}."
            )
        run_length = position - self._current_run.start
        self._current_run.close(run_length)
        self.runs.append(self._current_run)
        self._current_run = None

    def end(self, position: int) -> None:
        """Indicated the end of a sequence.

        After this function is called, the sequence is closed
        and is not possible to add more runs to this collection.

        Args:
            position (int): Position where the sequence ends.

        Raises:
            ValueError: Position is not equal to length of the sequence.
        """
        if self.is_run_open():
            self.close_run(position)
        self._is_ended = True

        if position != self.length:
            raise ValueError(
                f"Expected {self.length} base pairs in this sequence but processed {position}."
            )



import enum
import logging
import shlex
import subprocess
from time import sleep
from typing import Callable

from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.configuration import MANAGER_CFG, ExternalConfig, RepositoryConfig
from helix.progress.simple_worker import SimpleWorker
from helix.utility.external import External


class VariantCallingType(enum.Enum):
    InDel = enum.auto()
    SNP = enum.auto()
    Both = enum.auto()


class VariantCaller(SimpleWorker):
    """This class is responsible for calling variants for a given alignment-map file.

    Args:
        input (AlignmentMapFile):
            File that contains aligned reads.
        calling_type (VariantCallingType, optional):
            Type of variant calling to perform. Defaults to VariantCallingType.Both.
        repo_config (RepositoryConfig, optional):
            Configuration for reference genome repository. Defaults to
            MANAGER_CFG.REPOSITORY.
        ext_config (ExternalConfig, optional):
            Configuration for external tools. Defaults to MANAGER_CFG.EXTERNAL.
        external (External, optional):
            Object that contains functions to call external tools.
            Defaults to External().
        progress (Callable[[str, int], optional):
            Function that accept a status message and a percentage,
            for progress tracking. Defaults to None.
        logger (Logger, optional):
            Logger object. Defaults to logging.getLogger(__name__).

    Raises:
        RuntimeError:
            If the index stats inside `input` are not available.
        RuntimeError:
            If the reference genome for `input` is not available.
    """

    def __init__(
        self,
        input: AlignmentMapFile,
        calling_type: VariantCallingType = VariantCallingType.Both,
        repo_config: RepositoryConfig = MANAGER_CFG.REPOSITORY,
        ext_config: ExternalConfig = MANAGER_CFG.EXTERNAL,
        external: External = External(),
        progress: Callable[[str, int], None] = None,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:

        super().__init__(None)
        self._external = external
        self._ext_config = ext_config
        self._ploidy = str(repo_config.metadata.joinpath("ploidy.txt"))
        self._is_quitting = False
        self._quitting_ack = False
        self._current_operation = None
        self._progress = progress
        self._logger = logger
        self._input_file = input
        self._calling_type = calling_type

        if self._input_file.file_info.index_stats is None:
            raise RuntimeError("Index stats cannot be None for variant calling")
        if self._input_file.file_info.reference_genome.ready_reference is None:
            raise RuntimeError("Reference cannot be None for variant calling")

    def call(self):
        """Does exactly what you think it does."""
        return self.run()

    def run(self):
        self.current_file = self._input_file
        reference = str(
            self._input_file.file_info.reference_genome.ready_reference.fasta
        )
        self._logger.info(
            f"Calling variants with {self._input_file.file_info.reference_genome.ready_reference}"
        )
        input_file = self._input_file.path

        skip_variant_opt = ""
        if self._calling_type == VariantCallingType.InDel:
            skip_variant_opt = "-V snps"
        elif self._calling_type == VariantCallingType.SNP:
            skip_variant_opt = "-V indels"

        output_file = self._input_file.path.with_name(
            f"{self._input_file.path.stem}_{self._calling_type.name}.vcf.gz"
        )

        pileup_opt = (
            f"mpileup -B -I -C 50 --threads {self._ext_config.threads}"
            f'-f "{reference}" -Ou "{input_file}"'
        )
        call_opt = (
            f'call --ploidy-file "{self._ploidy}" {skip_variant_opt} -mv'
            f'-P 0 --threads {self._ext_config.threads} -Oz -o "{output_file}"'
        )
        tabix_opt = f'-p vcf "{output_file}"'

        pileup_opt = shlex.split(pileup_opt)
        call_opt = shlex.split(call_opt)
        tabix_opt = shlex.split(tabix_opt)

        progress = None
        if self._progress is not None:
            mapped_sequences = [
                x.mapped for x in self.current_file.file_info.index_stats
            ]
            mapped_sequences = sum(mapped_sequences)

            # 112 bytes seems to be the average bytes written for a single base
            total_bytes = (
                mapped_sequences
                * self.current_file.file_info.alignment_stats.average_length
                * 112
            )

            progress = BaseProgressCalculator(
                self._progress, total_bytes, "[1/2] Calling"
            )
            progress = progress.compute_on_write_bytes

        self._current_operation = self._external.bcftools(
            pileup_opt,
            stdout=subprocess.PIPE,
            io=progress,
        )

        call = self._external.bcftools(call_opt, stdin=self._current_operation.stdout)
        call.communicate()
        if self._is_quitting:
            self._quitting_ack = True
            return

        progress = BaseProgressCalculator(
            self._progress, output_file.stat().st_size, "[2/2] Indexing"
        )
        self._current_operation = self._external.tabix(
            tabix_opt, io=progress.compute_on_read_bytes
        )
        self._current_operation.wait()
        self._quitting_ack = True

    def kill(self):
        """Kill a variant calling operation."""
        self._is_quitting = True
        try:
            # The while is just to prevent a (unlikely IMO but who knows) situation
            # where _is_quitting is seen as False by the other thread after Popen.kill()
            # has been called.
            for _ in range(10):
                operation = self._current_operation
                if self._quitting_ack:
                    break
                if operation is not None:
                    operation.kill()
                self._current_operation = None
                if self._quitting_ack:
                    break
                sleep(0.5)
            if not self._quitting_ack:
                raise RuntimeError(
                    f"Failed 10 attempts to kill {self._current_operation}."
                )
        except Exception as e:
            self._logger.error(f"Error while killing variant calling: {e!s}")



from helix.data.tabular_data import TabularData


class Search:
    def __init__(self, data: dict[str, TabularData]) -> None:
        self.data = data

    def search(self, text):
        refined_data = {
            x: y for x, y in self.data.items() if x is not None and text in x
        }
        if len(refined_data) != len(self.data) and len(refined_data) != 0:
            return refined_data

        refined_data = {}
        for key, value in self.data.items():
            for row in value.rows:
                for col in row.columns:
                    if col and text in col:
                        refined_data[key] = value
                        break
        return refined_data



from math import floor, log, log10
from typing import OrderedDict


class UnitPrefix:
    """Convert a quantity into a more readable form using unit prefixes.

    This class is expecially useful for very big or very small number, that
    are difficult to read or visualize.

    Examples:
        >>> UnitPrefix.convert_bytes(1024)
        '1kB'
        >>> UnitPrefix.convert(1000)
        '1k'
        >>> UnitPrefix.convert(0.01)
        '10m'
    """

    def convert_bytes(input: int, decimal=2) -> str:
        """Convert bytes using binary prefixes.

        Args:
            input (int): The number to convert.
            decimal (int, optional): Number of decimals to use. Defaults to 2.

        Raises:
            ValueError: The input number is a float
            ValueError: The input number is negative.

        Returns:
            str: Converted number.
        """
        if input is None:
            return None
        if isinstance(input, float):
            raise ValueError("Can't convert a fractional number of bytes")
        elif input < 0:
            raise ValueError("Can't convert a negative number of bytes")
        converted = None
        si_prefixes = OrderedDict(
            [
                (" B", (0, 10, 0)),
                (" kB", (10, 20, 10)),
                (" MB", (20, 30, 20)),
                (" GB", (30, 40, 30)),
                (" TB", (40, None, 40)),
            ]
        )
        if input == 0:
            return "0 B"

        input_log = log(input, 2)
        input_log = int(floor(input_log))

        converted = None
        for letter, interval in si_prefixes.items():
            min_interval = interval[0]
            max_interval = interval[1]
            in_interval = False
            if max_interval is None:
                in_interval = input_log >= min_interval
            elif min_interval is None:
                in_interval = input_log < max_interval
            else:
                in_interval = input_log >= min_interval and input_log < max_interval

            if in_interval:
                converted = f"{(input / 2**interval[2]):.{decimal}f}"
                # Not interested in any .0000 at the end
                converted = converted.rstrip("0").rstrip(".")
                # If converted == "0" will become ""
                if converted == "":
                    converted = "0"
                converted += letter
                break
        return converted

    def convert(input: int, decimal=2) -> str:
        """Convert a quantity using metric prefixes.

        Args:
            input (int): The number to convert.
            decimal (int, optional): Number of decimals to use. Defaults to 2.

        Returns:
            str: Converted number.
        """
        converted = None
        si_prefixes = OrderedDict(
            [
                ("p", (None, -9, -12)),
                ("n", (-9, -6, -9)),
                ("μ", (-6, -3, -6)),
                ("m", (-3, 0, -3)),
                ("", (0, 3, 0)),
                ("k", (3, 6, 3)),
                ("M", (6, 9, 6)),
                ("G", (9, 12, 9)),
                ("T", (12, 15, 12)),
                ("P", (15, None, 15)),
            ]
        )
        if input == 0:
            return "0"

        input_log = log10(input)
        input_log = int(floor(input_log))

        converted = None
        for letter, interval in si_prefixes.items():
            min_interval = interval[0]
            max_interval = interval[1]
            in_interval = False
            if max_interval is None:
                in_interval = input_log >= min_interval
            elif min_interval is None:
                in_interval = input_log < max_interval
            else:
                in_interval = input_log >= min_interval and input_log < max_interval

            if in_interval:
                converted = f"{(input / 10**interval[2]):.{decimal}f}"
                # Not interested in any .0000 at the end
                converted = converted.rstrip("0").rstrip(".")
                # If converted == "0" will become ""
                if converted == "":
                    converted = "0"
                converted += letter
                break
        return converted



from pathlib import Path

from helix.reference.reference import Reference
from helix.utility.external import External


class MicroarrayGenerator:
    def __init__(
        self, vcf_file: Path, reference: Reference, external: External
    ) -> None:
        self.input = vcf_file
        self._external = external
        self._reference = reference

    def do(self):
        # output = ""
        # destination = ""
        # tabix_opt = f"-p vcf {self.input}"
        # bcftools_opt = f"annotate -Oz -a
        # {self._reference.ready_reference} -c CHROM,POS,ID {self.input}"
        # bcftools_opt = (
        # f"query -f '%ID\tCHROM\t%POS[\t%TGT\n' {output} -o {destination} "
        # )
        pass

    # VCF -> Indexing -> Annotate -> Indexing -> Query



class MockFileContent:
    def __init__(self, buffer) -> None:
        self.buffer = buffer
        self.seek = 0

    def __iter__(self):
        for line in self.buffer.split("\n"):
            yield line

    def write(self, buffer):
        self.buffer += buffer

    def read(self, b):
        max_read = min(len(self.buffer), self.seek + b)
        if (self.seek - max_read) == 0:
            return None
        self.seek = max_read
        return bytes(self.buffer[self.seek : max_read], "utf8")


class MockFile:
    def __init__(self, buffer) -> None:
        self.buffer = buffer

    def __enter__(self):
        return MockFileContent(self.buffer)

    def __exit__(self, _, _1, _2):
        return


class MockPath:
    class _StatObj:
        def __init__(self, size) -> None:
            self.st_size = size

    def __init__(self, path="foo", content={}, buffer="", exists=True):
        self.path = path
        self._content = content
        self._exists = exists
        self._buffer = buffer

    def joinpath(self, *str):
        target = "\\".join(str)
        if target in self._content:
            return self._content[target]
        raise ValueError()

    def exists(self):
        return self._exists

    def __str__(self) -> str:
        return self.path

    def rename(self, target):
        self.path = str(target)

    def open(self, *kargs, **kwargs):
        return MockFile(self._buffer)

    def stat(self):
        return MockPath._StatObj(len(self._buffer))

    def unlink(self):
        self._exists = False

    @property
    def name(self):
        return self.path

    @property
    def suffix(self):
        split = self.path.rsplit(".")
        if len(split) > 1:
            return split[1]
        return ""

    def with_suffix(self, suffix):
        no_suffix = self.path.rsplit(".")[0]
        target = no_suffix + "." + suffix
        if target in self._content:
            return self._content[target]
        return MockPath(target)



import filecmp
from pathlib import Path
from helix.configuration import RepositoryConfig
from helix.data.genome import Genome
from helix.files.downloader import Downloader
from test.genome_fixtures import remote_repo_fixture
from test.utility import MockPath


def test_downloaded_file_content_matches(remote_repo_fixture, tmp_path):
    cfg = RepositoryConfig()
    cfg.temporary = tmp_path.joinpath("temp")
    cfg.temporary.mkdir(exist_ok=True)

    downloader = Downloader(cfg)
    sut = downloader.perform(remote_repo_fixture["bgzip"]["genome"])

    assert filecmp.cmp(sut, remote_repo_fixture["bgzip"]["file_on_disk"])


def test_genome_md5_is_assigned_correctly(remote_repo_fixture, tmp_path):
    cfg = RepositoryConfig()
    cfg.temporary = tmp_path.joinpath("temp")
    cfg.temporary.mkdir(exist_ok=True)

    downloader = Downloader(cfg)
    sut = downloader.perform(remote_repo_fixture["bgzip"]["genome"])

    assert (
        remote_repo_fixture["bgzip"]["md5"]
        == remote_repo_fixture["bgzip"]["genome"].downloaded_md5
    )


def test_size_fetch_is_correct(remote_repo_fixture, tmp_path):
    cfg = RepositoryConfig()
    cfg.temporary = tmp_path.joinpath("temp")
    cfg.temporary.mkdir(exist_ok=True)

    downloader = Downloader(cfg)
    sut = downloader.perform(remote_repo_fixture["bgzip"]["genome"])

    assert (
        remote_repo_fixture["bgzip"]["size"]
        == remote_repo_fixture["bgzip"]["genome"].download_size
    )


def test_download_is_resumed_correctly(remote_repo_fixture, tmp_path):
    cfg = RepositoryConfig()
    cfg.temporary = tmp_path.joinpath("temp")
    cfg.temporary.mkdir(exist_ok=True)
    genome = remote_repo_fixture["bgzip"]["genome"]
    temporary_file = tmp_path.joinpath(genome.name_only)

    ten_bytes = remote_repo_fixture["bgzip"]["file_on_disk"].read_bytes()[0:10]
    temporary_file.write_bytes(ten_bytes)

    downloader = Downloader(cfg)
    sut = downloader.perform(remote_repo_fixture["bgzip"]["genome"])

    assert filecmp.cmp(sut, remote_repo_fixture["bgzip"]["file_on_disk"])


def test_already_downloaded_file_is_skipped(remote_repo_fixture, tmp_path):
    cfg = RepositoryConfig()
    cfg.temporary = tmp_path.joinpath("temp")
    cfg.temporary.mkdir(exist_ok=True)
    genome = remote_repo_fixture["bgzip"]["genome"]
    temporary_file = cfg.temporary.joinpath(genome.name_only)

    file = remote_repo_fixture["bgzip"]["file_on_disk"].read_bytes()
    temporary_file.write_bytes(file)
    last_edit = temporary_file.stat().st_mtime_ns

    downloader = Downloader(cfg)
    Path().stat().st_mtime_ns
    sut = downloader.perform(remote_repo_fixture["bgzip"]["genome"])

    # Ensure the file is not touched
    assert sut.stat().st_mtime_ns == last_edit



import hashlib
import logging
import os
from pathlib import Path

import certifi
import pycurl
from google.cloud import storage

from helix.configuration import MANAGER_CFG
from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.progress.file_io_monitor import FileIOMonitor
from helix.reference.genome_metadata_loader import Genome
from helix.files.file_type_checker import FileTypeChecker
from helix.utility.unit_prefix import UnitPrefix

HANDLERS = {}
SIZE_HANDLERS = {}


def uri_handler(uri):
    def decorator(f):
        global HANDLERS
        HANDLERS[uri] = f
        return f

    return decorator


def size_handler(uri):
    def decorator(f):
        global SIZE_HANDLERS
        SIZE_HANDLERS[uri] = f
        return f

    return decorator


class Downloader:

    def __init__(
        self,
        config=MANAGER_CFG.REPOSITORY,
        file_type_checker: FileTypeChecker = FileTypeChecker(),
        curl_class=pycurl.Curl,
    ) -> None:
        if file_type_checker is None:
            raise RuntimeError("FileTypeChecker cannot be None.")

        self._config = config
        self._curl_class = curl_class
        self.file_type_checker = file_type_checker
        self._logger = logging.getLogger("downloader")

    def size_pycurl(self, url: str):
        curl = self._curl_class()
        curl.setopt(pycurl.URL, url)
        curl.setopt(pycurl.NOBODY, 1)
        curl.perform()
        length = int(curl.getinfo(pycurl.CONTENT_LENGTH_DOWNLOAD))
        if length == -1:
            return None
        if length == 0:
            return None
        return length

    def get_file_size(self, url: str):
        for handler in SIZE_HANDLERS.keys():
            if url.startswith(handler):
                return SIZE_HANDLERS[handler](self, url)
        return self.size_pycurl(url)

    def calculate_md5_hash(self, filename: Path, chunk_size=4096):
        md5_hash = hashlib.md5()
        with filename.open("rb") as f:
            for chunk in iter(lambda: f.read(chunk_size), b""):
                md5_hash.update(chunk)
        return md5_hash.hexdigest()

    @size_handler("https://storage.cloud.google.com")
    @size_handler("gs://")
    def size_google(self, genome: Genome):
        storage_client = storage.Client.create_anonymous_client()
        uri = genome.fasta_url.strip()
        if not uri.startswith("gs://"):
            uri = "gs://" + uri.replace("https://storage.cloud.google.com/", "")

        blob = storage.Blob.from_string(uri, client=storage_client)
        blob.reload()
        return blob.size

    @uri_handler("https://storage.cloud.google.com")
    @uri_handler("gs://")
    def download_google(self, genome: Genome, callback: any):
        storage_client = storage.Client.create_anonymous_client()
        uri = genome.fasta_url.strip()
        if not uri.startswith("gs://"):
            uri = "gs://" + uri.replace("https://storage.cloud.google.com/", "")

        blob = storage.Blob.from_string(uri, client=storage_client)
        blob.reload()
        if genome.download_size is None:
            genome.download_size = blob.size
        if genome.downloaded_md5 is None:
            genome.downloaded_md5 = blob.md5_hash

        target = self._config.temporary.joinpath(genome.name_only)

        if target.exists():
            if target.stat().st_size == genome.download_size:
                return self.post_download_action(genome, target)
        base_calc = BaseProgressCalculator(callback, genome.download_size, "Download")
        monitor = FileIOMonitor(
            target, base_calc.compute_on_write_bytes, genome.download_size
        )
        blob.download_to_filename(target)
        monitor.quit()
        return self.post_download_action(genome, target)

    def post_download_action(self, genome: Genome, downloaded: Path):
        md5 = self.calculate_md5_hash(downloaded)
        if genome.downloaded_md5 is None:
            genome.downloaded_md5 = md5
        elif genome.downloaded_md5 != md5:
            downloaded.unlink()
            raise RuntimeError(f"MD5 for {genome} is not matching and was deleted.")

        extension = self.file_type_checker.get_extension(downloaded)

        if extension is None:
            _, ext = os.path.splitext(genome.fasta_url)
            extension = ext.lower()

        if extension is None:
            raise RuntimeError(
                f"Format file for {genome} was not recognized. Aborting."
            )

        if downloaded.suffix != extension:
            target = downloaded.with_suffix(extension)
            if target.exists():
                target.unlink()
            downloaded.rename(target)
            downloaded = target
        return downloaded

    def perform(self, genome: Genome, progress: any = None) -> Path:
        for handler in HANDLERS.keys():
            if genome.fasta_url.startswith(handler):
                return HANDLERS[handler](self, genome, progress)
        return self.download_pycurl(genome, progress)

    def download_pycurl(self, genome: Genome, progress: any):
        if genome.download_size is None:
            genome.download_size = self.get_file_size(genome.fasta_url)
            if genome.download_size is None:
                raise RuntimeError(f"Unable to download file {genome.fasta_url}")

        target = self._config.temporary.joinpath(genome.name_only)

        resume_from = None
        if target is not None and target.exists():
            if target.stat().st_size == genome.download_size:
                return self.post_download_action(genome, target)
            else:
                resume_from = target.stat().st_size

        progress_calc = None
        if progress is not None:
            progress_calc = BaseProgressCalculator(
                progress, genome.download_size, "Download"
            )

        total = UnitPrefix.convert_bytes(genome.download_size)
        resume = UnitPrefix.convert_bytes(resume_from if resume_from is not None else 0)
        url = genome.fasta_url
        self._logger.info(
            "Downloading file from %s, resuming from %s (%s total)"
            % (url, resume, total)
        )

        self.download_file(genome.fasta_url, target, resume_from, progress_calc)
        return self.post_download_action(genome, target)

    def download_file(
        self,
        url: str,
        target: Path,
        resume_from: int = None,
        progress_calc: BaseProgressCalculator = None,
    ):
        curl = self._curl_class()
        curl.setopt(pycurl.URL, url)
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        curl.setopt(pycurl.CAINFO, certifi.where())

        if resume_from is not None:
            curl.setopt(pycurl.RESUME_FROM, resume_from)

        if progress_calc is not None:
            curl.setopt(pycurl.NOPROGRESS, False)
            curl.setopt(
                pycurl.PROGRESSFUNCTION,
                lambda _tot_down, down, _tot_up, _up: progress_calc.compute(down),
            )

        try:
            with target.open("wb" if resume_from is None else "ab") as f:
                curl.setopt(pycurl.WRITEDATA, f)
                curl.perform()
                curl.close()
        finally:
            if progress_calc is not None:
                progress_calc.compute(None)



from test.utility import MockFile, MockPath

import pytest

from helix.fasta.fasta_letter_counter import FASTALetterCounter


class MockGenome:
    def __init__(self, fasta, dict) -> None:
        self.fasta = fasta
        self.dict = dict


def gzip_open(path, mode):
    return MockFile(path._buffer)


def test_no_dictionary():
    fa_lines = f">1 irrelevant\n{'N'*5}\n{'N'*5}\n>2 irrelevant\n"

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(buffer=fa_lines), MockPath(exists=False))
        sut = FASTALetterCounter(genome)
        assert "Unable to find dictionary" in str(e.value)


def test_sequence_not_in_dictionary():
    fa_lines = f">1 irrelevant\n{'N'*5}\n{'N'*5}\n>2 irrelevant\n"
    dict_lines = (
        "@HD\tVN:1.0\tSO:unsorted\n@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz"
    )
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(ValueError) as e:
        genome = MockGenome(MockPath(buffer=fa_lines), MockPath(buffer=dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "not present in dictionary" in str(e.value)


def test_fastq():
    fa_lines = ">1 irrelevant\n{'N'*5}\n{'N'*5}\n+2 irrelevant\n"
    dict_lines = (
        "@HD\tVN:1.0\tSO:unsorted\n@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz"
    )
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(buffer=fa_lines), MockPath(buffer=dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "Expected a FASTA" in str(e.value)


def test_duplicate_sequence():
    fa_lines = f">1 irrelevant\n{'N'*5}\n{'N'*5}\n>1 irrelevant\n"
    dict_lines = (
        "@HD\tVN:1.0\tSO:unsorted\n@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz"
    )
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(buffer=fa_lines), MockPath(buffer=dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "duplicated sequence" in str(e.value)


def test_only_comments():
    fa_lines = f"#Hello\n#Foo\n"
    dict_lines = (
        "@HD\tVN:1.0\tSO:unsorted\n@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz"
    )
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(buffer=fa_lines), MockPath(buffer=dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "no sequences" in str(e.value)



from helix.data.genome import Genome
from helix.data.tabular_data import TabularData, TabularDataRow
from helix.utility.unit_prefix import UnitPrefix


class GenomeAdapter:
    def adapt(genome: Genome):
        column = {
            "FASTA URL": genome.fasta_url,
            "FAI URL": genome.fai_url,
            "GZI URL": genome.gzi_url,
            "Suffix": genome.suffix,
            "Build": genome.build,
            "Source": genome.source,
            "Sequences": (
                "Available" if genome.sequences is not None else "Not available"
            ),
            "Description": genome.description,
            "Size": UnitPrefix.convert_bytes(genome.download_size),
            "Decompressed Size": UnitPrefix.convert_bytes(genome.decompressed_size),
            "BGZip Size": UnitPrefix.convert_bytes(genome.bgzip_size),
            "MD5": genome.downloaded_md5,
            "Decompressed MD5": genome.decompressed_md5,
            "BGZip MD5": genome.bgzip_md5,
            "MtDNA Model": genome.mitochondrial_model,
            "Parent Folder": genome.parent_folder,
        }
        return TabularData(
            None, [TabularDataRow(x, [str(y)]) for x, y in column.items()]
        )



import gzip
import logging
from pathlib import Path
from typing import Optional

from parse import parse
from pydantic import BaseModel

from helix.naming.converter import Converter


class RawEntry(BaseModel, frozen=True):
    id: str
    chromosome: str
    position: int
    result: Optional[str] = None

    @property
    def template_entry(self):
        return f"{self.id}\t{self.chromosome}\t{self.position}\n"

    def __str__(self) -> str:
        return f"{self.id},{self.chromosome},{self.position},{self.result}"

    def __repr__(self) -> str:
        return self.__str__()


class MicroarrayMeta(BaseModel):
    input_format: str = "{id}\t{chromosome}\t{position}\n"
    output_format: str = "{id}\t{chromosome}\t{position}\t{result}\n"
    file_extension: Optional[str] = None
    undetermined: str = "--"
    skip: int = 0


class ParsedMicroarrayFile(BaseModel):
    grouped_entries: dict[str, dict[int, set[RawEntry]]]
    comments: Optional[list[str]] = None
    meta: Optional[MicroarrayMeta] = None


class RawFile:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.meta: MicroarrayMeta = MicroarrayMeta()
        self._logger = logging.getLogger(__name__)
        if self.path.suffix == ".gz":
            self.open_file = lambda: gzip.open(self.path, "rt")
        else:
            self.open_file = lambda: self.path.open("rt")

    def load(self) -> ParsedMicroarrayFile:
        grouped: dict[str, dict[int, set[RawEntry]]] = dict()
        meta: MicroarrayMeta = None
        comments = []

        # Comments are only at the beginning of the file: Parse the
        # comments and do another cycle to to parse the rest to avoid
        # the overhead of these intial if line.startswith() which are
        # REALLY slow (Python 3.12).
        with self.open_file() as file:
            for line in file:
                if line.startswith("##"):
                    line = line.removeprefix("##")
                    if meta is not None:
                        self._logger.warning(
                            f"Duplicate metadata line found in {self.path!s}"
                        )
                    meta = MicroarrayMeta.model_validate_json(line)
                    continue
                elif line.startswith("#"):
                    comments.append(line)
                    continue
                else:
                    break
            canonicalized = {}
            index = 0
            # If no metadata was found, just create a default object
            if meta is None:
                meta = MicroarrayMeta()

            for line in file:
                if index < meta.skip:
                    index += 1
                    continue
                parsed = parse(self.meta.output_format, line)
                if parsed is None:
                    parsed = parse(self.meta.input_format, line)
                    if parsed is None:
                        self._logger.warning(
                            f"Found invalid line in {self.path!s}: {self.meta.input_format}!={line}"
                        )
                        continue

                entry = RawEntry(
                    id=parsed["id"],
                    chromosome=parsed["chromosome"],
                    position=parsed["position"],
                    result=parsed["result"] if "result" in parsed else None,
                )
                if entry.chromosome not in canonicalized:
                    canonicalized[entry.chromosome] = Converter.canonicalize(
                        entry.chromosome
                    )
                chromosome = canonicalized[entry.chromosome]
                if chromosome not in grouped:
                    grouped[chromosome] = dict()
                if entry.position not in grouped[chromosome]:
                    grouped[chromosome][entry.position] = set()
                grouped[chromosome][entry.position].add(entry)

        if meta.file_extension is None:
            if len(self.path.suffixes) > 1 and self.path.suffixes[-1] == ".gz":
                meta.file_extension = self.path.suffixes[-2]
            elif len(self.path.suffixes) > 0:
                meta.file_extension = self.path.suffixes[-1]

        return ParsedMicroarrayFile(
            grouped_entries=grouped,
            comments=comments,
            meta=meta,
        )



from enum import Flag, auto
import uuid

from helix.configuration import MANAGER_CFG
from helix.data.sequence import Sequence
from helix.naming.converter import Converter


class RegionType(Flag):
    Y = auto()
    MT = auto()
    WES = auto()


class Regions:
    """Represent a region in a genome.

    This class loads and manipulate .bed file templates containing regions in a genome.
    Since the bed files need to change according to the specific sequence naming
    of the loaded file, this class take care of loading the templates for these
    .bed files and convert them to the correct sequence naming.
    """

    def __init__(self, config=MANAGER_CFG.REPOSITORY, converter=Converter()) -> None:
        self._templates = config.metadata.joinpath("bed_templates")
        self._temporary = config.temporary
        self._converter = converter

    def get_path(self, build: str, type: RegionType, sequences: list[Sequence]):
        """Return the path of a .bed file that is built for a specific genome.

        Args:
            build (str): The build number (currently only 37 or 38 are supported)
            type (RegionType): RegionType of the region the caller is looking for.
                Currently only Y or MT or WES are supported. Regions can be combined
                but only Y | MT is a valid combination.
            sequences (list[Sequence]): List of sequences from the input file.

        Raises:
            ValueError: An invalid combination of RegionTypes was provided.
            FileNotFoundError: Template file for the provided build was not found.

        Returns:
            pathlib.Path: Path to the.bed file that is built for a specific genome.
        """
        name_map = {x.canonic_name: x.name for x in sequences}

        suffix = ""
        if type & RegionType.Y and type & RegionType.WES:
            raise ValueError("Invalid combination: Y and WES")
        if type & RegionType.MT and type & RegionType.WES:
            raise ValueError("Invalid combination: MT and WES")

        if type & RegionType.Y:
            suffix += "_y"
        if type & RegionType.MT:
            suffix += "_mt"
        if type & RegionType.WES:
            suffix += "_wes"
        file_name = f"{build}{suffix}.csv"
        file = self._templates.joinpath(file_name)
        if not file.exists():
            raise FileNotFoundError(f"Unable to find BED template at: {file!s}")
        with file.open("r") as f:
            lines = f.readlines()
            rows = []
            for line in lines:
                line = line.split(",")
                sequence = line[0]
                begin = line[1]
                end = line[2].strip()
                if sequence in name_map:
                    sequence = name_map[sequence]
                rows.append(f"{sequence}\t{begin}\t{end}\n")
        temporary_file = self._temporary.joinpath(str(uuid.uuid4()))
        with temporary_file.open("w") as f:
            f.writelines(rows)
        return temporary_file

    def get_content(self, build: str, type: RegionType):
        """Build and get the content of a .bed file.

        Args:
            build (str): Build name.
            type (RegionType): RegionType of the region.

        Returns:
            str: Content of the .bed file.
        """
        file = self.get_path(build, type)
        return file.read_text("utf-8")



import base64
from pathlib import Path
import sys

if len(sys.argv) < 2:
    exit()
path = Path(sys.argv[1])
with path.open("rb") as f:
    b = f.read()
    print(f'"{path.name}" : "{base64.standard_b64encode(b).decode("utf-8")}"')



import gzip
import hashlib
import logging
import shutil
import typing
import zipfile
from pathlib import Path

from helix.files.gzip import GZip, GZipAction
from helix.reference.genome_metadata_loader import Genome
from helix.files.file_type_checker import FileType, FileTypeChecker


class Decompressor:
    def __init__(
        self,
        type_checker: FileTypeChecker = FileTypeChecker(),
        gzip_compressor=GZip(),
    ) -> None:
        self._gzip_compressor = gzip_compressor
        self._type_checker = type_checker

        self._handlers: typing.Dict[FileType, typing.Callable[[Path, Path], None]] = {
            FileType.GZIP: Decompressor.razf_gzip,
            FileType.RAZF_GZIP: Decompressor.razf_gzip,
            FileType.ZIP: Decompressor.zip,
            FileType.SEVENZIP: Decompressor.sevenzip,
            FileType.BZIP: Decompressor.bzip,
            FileType.DECOMPRESSED: Decompressor.dummy,
            FileType.BGZIP: Decompressor.dummy,
        }

    def dummy(self, input_file: Path, output_file: Path):
        # Dummy handler as the file it's either already compressed
        # in the target format or decompressed.
        if input_file != output_file:
            if output_file.exists():
                output_file.unlink()
            input_file.rename(output_file)

    def gz(self, input_file: Path, output_file: Path):
        # Not reliable with RAZF and currently not used.
        with gzip.open(str(input_file), "rb") as f_in:
            with open(output_file, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

    def sevenzip(self, input_file: Path, output_file: Path):
        raise NotImplementedError()

    def bzip(self, input_file: Path, output_file: Path):
        raise NotImplementedError()

    def zip(self, input_file: Path, output_file: Path):
        with zipfile.ZipFile(str(input_file), "r") as f:
            files = f.namelist()
            if len(files) > 1:
                raise RuntimeError(
                    f"Error decompressing {input_file!s}: zip contains more than 1 file"
                )
            extracted = Path(f.extract(files[0], output_file.parent))
            if extracted != output_file:
                extracted.rename(output_file)

    def razf_gzip(self, input_file: Path, output_file: Path):
        self._gzip_compressor.gzip(input_file, output_file, GZipAction.Decompress)

    def calculate_md5_hash(self, filename: Path, chunk_size=4096):
        md5_hash = hashlib.md5()
        with filename.open("rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                md5_hash.update(chunk)
        return md5_hash.hexdigest()

    def perform(self, genome: Genome, downloaded: Path = None):
        if not downloaded.exists():
            raise FileNotFoundError(
                f"Error decompressing {str(genome)}: "
                f"unable to find input file {downloaded!s}"
            )

        no_ext = downloaded.name.removesuffix("".join(downloaded.suffixes))
        target = downloaded.with_name(no_ext + ".fa")
        type = self._type_checker.get_type(downloaded)
        if type not in self._handlers:
            raise RuntimeError(f"Error decompressing {str(genome)}: unknown type")

        handler = self._handlers[type]
        logging.debug(
            f"Decompressing {downloaded!s}. {type.name} compression detected."
        )
        handler(self, downloaded, target)
        if not target.exists():
            raise RuntimeError(
                f"Error decompressing {str(genome)}: Decompressed file not found"
            )

        type = self._type_checker.get_type(target)

        if type != FileType.DECOMPRESSED:
            # If we've a bgzip, the file is never decompressed.
            return target

        if genome.decompressed_size is None:
            genome.decompressed_size = target.stat().st_size
        elif genome.decompressed_size != target.stat().st_size:
            raise RuntimeError(f"Error decompressing {str(genome)}: size mismatch")

        if genome.decompressed_md5 is None:
            genome.decompressed_md5 = self.calculate_md5_hash(target)
        elif genome.decompressed_md5 != self.calculate_md5_hash(target):
            raise RuntimeError(f"Error decompressing {str(genome)}: MD5 mismatch")

        return target



import logging
from pathlib import Path
import time
from threading import Thread

try:
    import debugpy
except ImportError:
    pass


class FileIOMonitor(Thread):
    def __init__(self, path: Path, report, target, polling=1) -> None:
        super().__init__()
        self.path: Path = path
        self.target = target
        self.report = report
        self.polling = polling
        self._quitting = False
        self.start()

    def run(self):
        if debugpy:
            if debugpy.is_client_connected():
                debugpy.debug_this_thread()
        try:
            while not self._quitting:
                write_bytes = self.path.stat().st_size
                self.report(None, write_bytes)
                if write_bytes == self.target:
                    break
                time.sleep(self.polling)
        except Exception as e:
            # The process can end anytime and psutil will fail.
            # That's not an error.
            logging.error(f"Exception in FileIOMonitor: {e!s}")
            pass
        finally:
            # Signal the process ended.
            self.report(None, None)

    def quit(self):
        self._quitting = True



from helix.utility.unit_prefix import UnitPrefix


def test_base_conversion():
    assert UnitPrefix.convert(10**-15, 3) == "0.001p"
    assert UnitPrefix.convert(10**-14, 2) == "0.01p"
    assert UnitPrefix.convert(10**-13) == "0.1p"
    assert UnitPrefix.convert(10**-12) == "1p"
    assert UnitPrefix.convert(10**-11) == "10p"
    assert UnitPrefix.convert(10**-10) == "100p"
    assert UnitPrefix.convert(10**-9) == "1n"
    assert UnitPrefix.convert(10**-8) == "10n"
    assert UnitPrefix.convert(10**-7) == "100n"
    assert UnitPrefix.convert(10**-6) == "1μ"
    assert UnitPrefix.convert(10**-5) == "10μ"
    assert UnitPrefix.convert(10**-4) == "100μ"
    assert UnitPrefix.convert(10**-3) == "1m"
    assert UnitPrefix.convert(10**-2) == "10m"
    assert UnitPrefix.convert(10**-1) == "100m"
    assert UnitPrefix.convert(0) == "0"
    assert UnitPrefix.convert(10**0) == "1"
    assert UnitPrefix.convert(10**1) == "10"
    assert UnitPrefix.convert(10**2) == "100"
    assert UnitPrefix.convert(10**3) == "1k"
    assert UnitPrefix.convert(10**4) == "10k"
    assert UnitPrefix.convert(10**5) == "100k"
    assert UnitPrefix.convert(10**6) == "1M"
    assert UnitPrefix.convert(10**7) == "10M"
    assert UnitPrefix.convert(10**8) == "100M"
    assert UnitPrefix.convert(10**9) == "1G"
    assert UnitPrefix.convert(10**10) == "10G"
    assert UnitPrefix.convert(10**11) == "100G"
    assert UnitPrefix.convert(10**12) == "1T"
    assert UnitPrefix.convert(10**13) == "10T"
    assert UnitPrefix.convert(10**14) == "100T"
    assert UnitPrefix.convert(10**15) == "1P"
    assert UnitPrefix.convert(10**16) == "10P"
    assert UnitPrefix.convert(10**17) == "100P"
    assert UnitPrefix.convert(10**18) == "1000P"
    assert UnitPrefix.convert(10**19) == "10000P"


def test_bytes_conversion():
    assert UnitPrefix.convert_bytes(0) == "0 B"
    assert UnitPrefix.convert_bytes(2**0) == "1 B"
    assert UnitPrefix.convert_bytes(2**1) == "2 B"
    assert UnitPrefix.convert_bytes(2**2) == "4 B"
    assert UnitPrefix.convert_bytes(2**3) == "8 B"
    assert UnitPrefix.convert_bytes(2**4) == "16 B"
    assert UnitPrefix.convert_bytes(2**5) == "32 B"
    assert UnitPrefix.convert_bytes(2**6) == "64 B"
    assert UnitPrefix.convert_bytes(2**7) == "128 B"
    assert UnitPrefix.convert_bytes(2**8) == "256 B"
    assert UnitPrefix.convert_bytes(2**9) == "512 B"
    assert UnitPrefix.convert_bytes(2**10) == "1 kB"
    assert UnitPrefix.convert_bytes(2**11) == "2 kB"
    assert UnitPrefix.convert_bytes(2**12) == "4 kB"
    assert UnitPrefix.convert_bytes(2**13) == "8 kB"
    assert UnitPrefix.convert_bytes(2**14) == "16 kB"
    assert UnitPrefix.convert_bytes(2**15) == "32 kB"
    assert UnitPrefix.convert_bytes(2**16) == "64 kB"
    assert UnitPrefix.convert_bytes(2**17) == "128 kB"
    assert UnitPrefix.convert_bytes(2**18) == "256 kB"
    assert UnitPrefix.convert_bytes(2**19) == "512 kB"
    assert UnitPrefix.convert_bytes(2**20) == "1 MB"
    assert UnitPrefix.convert_bytes(2**21) == "2 MB"
    assert UnitPrefix.convert_bytes(2**22) == "4 MB"
    assert UnitPrefix.convert_bytes(2**23) == "8 MB"
    assert UnitPrefix.convert_bytes(2**24) == "16 MB"
    assert UnitPrefix.convert_bytes(2**25) == "32 MB"
    assert UnitPrefix.convert_bytes(2**26) == "64 MB"
    assert UnitPrefix.convert_bytes(2**27) == "128 MB"
    assert UnitPrefix.convert_bytes(2**28) == "256 MB"
    assert UnitPrefix.convert_bytes(2**29) == "512 MB"
    assert UnitPrefix.convert_bytes(2**30) == "1 GB"
    assert UnitPrefix.convert_bytes(2**31) == "2 GB"
    assert UnitPrefix.convert_bytes(2**32) == "4 GB"
    assert UnitPrefix.convert_bytes(2**33) == "8 GB"
    assert UnitPrefix.convert_bytes(2**34) == "16 GB"
    assert UnitPrefix.convert_bytes(2**35) == "32 GB"
    assert UnitPrefix.convert_bytes(2**36) == "64 GB"
    assert UnitPrefix.convert_bytes(2**37) == "128 GB"
    assert UnitPrefix.convert_bytes(2**38) == "256 GB"
    assert UnitPrefix.convert_bytes(2**39) == "512 GB"
    assert UnitPrefix.convert_bytes(2**40) == "1 TB"
    assert UnitPrefix.convert_bytes(2**41) == "2 TB"
    assert UnitPrefix.convert_bytes(2**42) == "4 TB"
    assert UnitPrefix.convert_bytes(2**43) == "8 TB"
    assert UnitPrefix.convert_bytes(2**44) == "16 TB"
    assert UnitPrefix.convert_bytes(2**45) == "32 TB"
    assert UnitPrefix.convert_bytes(2**46) == "64 TB"
    assert UnitPrefix.convert_bytes(2**47) == "128 TB"
    assert UnitPrefix.convert_bytes(2**48) == "256 TB"
    assert UnitPrefix.convert_bytes(2**49) == "512 TB"
    assert UnitPrefix.convert_bytes(2**50) == "1024 TB"



import logging
import os
import signal
from subprocess import Popen
from threading import Thread

import debugpy


class SimpleWorker(Thread):
    """Execute a function in another thread.

    It can work in two ways:
        1. Supplying a function that returns a Popen object
        2. Supplying any function and an object that contains a
           kill function
    Example:
        >>> worker = SimpleWorker(None, lambda: Popen(["sleep", "10"]))
        >>> worker.kill()
    In this case SimpleWorker will call the lambda, store the Popen object
    and call .communicate() on it.
        >>> worker = SimpleWorker(MyObjectClass, my_object.long_operation)
        >>> worker.kill()
    In this case SimpleWorker will call long_operation in another thread
    and call my_object.kill() when attempting to kill the operation. How the
    operation will be killed is delegated to MyObjectClass.
    """

    def __init__(self, object, function, *args, **kwargs) -> None:
        super().__init__()
        self.function = function
        self.object = object
        self._kwargs = kwargs
        self._args = args
        if self.function is not None:
            self.start()

    def run(self):
        if debugpy.is_client_connected():
            debugpy.debug_this_thread()
        self.object = self.function(*self._args, **self._kwargs)
        if self.object is not None and isinstance(self.object, Popen):
            self.object.communicate()

    def kill(self):
        object = self.object
        if object is None:
            logging.error("Trying to kill a SimpleWorker but there's nothing to kill.")
            return
        try:
            if hasattr(os.sys, "winver") or not isinstance(object, Popen):
                object.kill()
            else:
                object.send_signal(signal.SIGTERM)
        except Exception as e:
            logging.error(f"Error when killing process {e!s}")



from helix.data.coverage_stats import CoverageStats, DepthBin
from helix.data.tabular_data import TabularData, TabularDataRow
from helix.utility.unit_prefix import UnitPrefix


class CoverageStatsAdapter:
    def adapt(input_data: list[CoverageStats]):
        rows = []
        for sequence_stats in input_data:
            row = [
                sequence_stats.sequence_name,
                str(UnitPrefix.convert(sequence_stats.bin_entries[DepthBin.Zero])),
                str(
                    UnitPrefix.convert(
                        sequence_stats.bin_entries[DepthBin.Between0And3]
                    )
                ),
                str(
                    UnitPrefix.convert(
                        sequence_stats.bin_entries[DepthBin.Between3And7]
                    )
                ),
                str(UnitPrefix.convert(sequence_stats.bin_entries[DepthBin.MoreThan7])),
                str(UnitPrefix.convert(sequence_stats.all_average)),
                str(UnitPrefix.convert(sequence_stats.non_zero_average)),
            ]
            rows.append(TabularDataRow(None, row))
        horizontal_header = [
            "Sequence Name",
            "# Reads: 0",
            "# Reads: 0-3",
            "# Reads: 3-7",
            "# Reads: >7",
            "Average",
            "Non-zero average",
        ]
        return TabularData(horizontal_header, rows)



import filecmp
import shutil
from helix.data.genome import Genome
from helix.files.decompressor import Decompressor
from test.genome_fixtures import remote_repo_fixture, format_map


def test_decompress_gzip(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["gzip"]
    genome: Genome = remote_repo_fixture["gzip"]["genome"]
    shutil.copy(remote_repo_fixture["gzip"]["file_on_disk"], target)
    sut = Decompressor()
    output = sut.perform(genome, target)
    assert filecmp.cmp(output, remote_repo_fixture["fasta"]["file_on_disk"])
    assert remote_repo_fixture["fasta"]["md5"] == genome.decompressed_md5
    assert remote_repo_fixture["fasta"]["size"] == genome.decompressed_size


def test_decompress_fasta(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["fasta"]
    genome: Genome = remote_repo_fixture["fasta"]["genome"]
    shutil.copy(remote_repo_fixture["fasta"]["file_on_disk"], target)
    sut = Decompressor()
    output = sut.perform(genome, target)
    assert filecmp.cmp(output, remote_repo_fixture["fasta"]["file_on_disk"])
    assert remote_repo_fixture["fasta"]["md5"] == genome.decompressed_md5
    assert remote_repo_fixture["fasta"]["size"] == genome.decompressed_size


def test_decompress_bgzip(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["bgzip"]
    genome: Genome = remote_repo_fixture["bgzip"]["genome"]
    shutil.copy(remote_repo_fixture["bgzip"]["file_on_disk"], target)
    d = Decompressor()
    sut = d.perform(genome, target)

    # Testing the shortcut: should not touch a bgzip file
    assert filecmp.cmp(sut, remote_repo_fixture["bgzip"]["file_on_disk"])
    assert None == genome.decompressed_md5
    assert None == genome.decompressed_size


def test_decompress_zip(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["zip"]
    genome: Genome = remote_repo_fixture["zip"]["genome"]
    shutil.copy(remote_repo_fixture["zip"]["file_on_disk"], target)
    sut = Decompressor()
    output = sut.perform(genome, target)
    assert filecmp.cmp(output, remote_repo_fixture["fasta"]["file_on_disk"])
    assert remote_repo_fixture["fasta"]["md5"] == genome.decompressed_md5
    assert remote_repo_fixture["fasta"]["size"] == genome.decompressed_size


def test_decompress_razf(remote_repo_fixture, tmp_path_factory):
    target = tmp_path_factory.mktemp("tmp") / format_map["razf"]
    genome: Genome = remote_repo_fixture["razf"]["genome"]
    shutil.copy(remote_repo_fixture["razf"]["file_on_disk"], target)
    sut = Decompressor()
    output = sut.perform(genome, target)
    assert filecmp.cmp(output, remote_repo_fixture["fasta"]["file_on_disk"])
    assert remote_repo_fixture["fasta"]["md5"] == genome.decompressed_md5
    assert remote_repo_fixture["fasta"]["size"] == genome.decompressed_size



from pathlib import Path

import pytest

from helix.data.genome import Genome
import base64

# Each contains the same reference in different formats with a single sequence named chr10 having 60bp long.
format_map = {
    "bgzip": "fake_genome_bgzip.fasta.gz",
    "gzip": "fake_genome.fasta.gz",
    "fasta": "fake_genome.fasta",
    "zip": "fake_genome.zip",
    "razf": "fake_genome_razf.fasta.gz",
}
reverse_format_map = {x: y for y, x in format_map.items()}

file_map = {
    format_map[
        "bgzip"
    ]: "H4sIBAAAAAAA/wYAQkMCAE8As0vOKDI04OVyD3F2d3R0d3R2dwfSIY4hIc5A6Oju7AgUBTKcQSKOzs4gDJIHiTsDhdydHUMAtSygWkQAAAAfiwgEAAAAAAD/BgBCQwIAGwADAAAAAAAAAAAA",
    format_map[
        "gzip"
    ]: "H4sICKGhYmYEAGZha2VfZ2Vub21lLmZhc3RhABWKsQkAMQAC+4ffJVkgIBYu4AZpUmf/IoqIcrr2uXP8n0wBAqWkYTOCiNAUloCsu5czKAc/tSygWkQAAAA=",
    format_map[
        "fasta"
    ]: "PmNocjEwDQpHVENHQUFHQUNHR0dBQVRBVFRDVENUQUdDQUFBR0NUQUNBVFRDQUNDQ0FDQ0FBVEFDQUFBQ1RUQ0dDQVQ=",
    format_map[
        "zip"
    ]: "UEsDBBQAAAAIAF0/x1i1LKBaNQAAAEQAAAARAAAAZmFrZV9nZW5vbWUuZmFzdGEVibEJADEAAvuH3yVZICAWLuAGaVJn/yKKiHK39rlz/J9MAQKlrGEzgYjQHJaAbOvLGaSIB1BLAQI/ABQAAAAIAF0/x1i1LKBaNQAAAEQAAAARACQAAAAAAAAAIAAAAAAAAABmYWtlX2dlbm9tZS5mYXN0YQoAIAAAAAAAAQAYAOIsJMmfuNoB0n99zZ+42gFLn4mpn7jaAVBLBQYAAAAAAQABAGMAAABkAAAAAAA=",
    format_map[
        "razf"
    ]: "H4sIBAAAAAAAAwcAUkFaRgGAALNLzigyNODlcg9xdnd0dHd0dncH0iGOISHOQOjo7uwIFAUynEEijs7OIAySB4k7A4WACkIAtSygWkQAAAAAAAAAAAAAAAAAAAAAAAAAAAAARAAAAAAAAABQ",
}

md5_map = {
    format_map["bgzip"]: "ca76b4375bd41f81cfe2003d314dd4b2",
    format_map["gzip"]: "a070ccf700131775e1d8a3cc9128107a",
    format_map["fasta"]: "069c8ead795424e20e4a21fc5e368599",
    format_map["zip"]: "c463c4be4bf8d9a03bcd99ab74eaf1b8",
    format_map["razf"]: "4e299205f6a94ed9b18795f81fa80b33",
}


@pytest.fixture()
def remote_repo_fixture(tmp_path: Path):
    genomes = {}
    tmp_path = tmp_path.joinpath("repo")
    tmp_path.mkdir(exist_ok=True)
    for name, content in file_map.items():
        path = tmp_path.joinpath(name).absolute()
        with path.open("wb") as f:
            f.write(base64.b64decode(content))
        uri_path = path.as_uri()  # e.g., file://c:/.. or file://home/..

        genome = Genome(uri_path, build="38", source="Fake")
        genomes[reverse_format_map[name]] = {
            "genome": genome,
            "file_on_disk": path,
            "md5": md5_map[name],
            "size": path.stat().st_size,
        }
    return genomes



from collections import OrderedDict
from helix.data.alignment_map.alignment_map_file_info import AlignmentMapFileInfo
from helix.data.tabular_data import TabularData, TabularDataRow
from helix.reference.reference import ReferenceStatus
from helix.utility.unit_prefix import UnitPrefix


class AlignmentMapFileInfoAdapter:
    def adapt(stats: AlignmentMapFileInfo) -> TabularData:
        label_map = OrderedDict(
            [
                ("sorted", "Sorted"),
                ("indexed", "Indexed"),
                ("file_type", "File type"),
                ("content", "Content"),
                ("gender", "Gender"),
            ]
        )

        data = OrderedDict()
        data["Directory"] = [str(stats.path.parent)]
        data["Filename"] = [stats.path.name]
        data["Size"] = [UnitPrefix.convert_bytes(stats.path.stat().st_size)]
        data["File type"] = [stats.file_type.name]
        if stats.reference_genome.status == ReferenceStatus.Available:
            data["Reference"] = [
                f"Based on GRCh{stats.reference_genome.build}, available"
            ]
        elif stats.reference_genome.status == ReferenceStatus.Buildable:
            data["Reference"] = [
                f"Likely based on GRCh{stats.reference_genome.build}, buildable"
            ]
        elif stats.reference_genome.status == ReferenceStatus.Downloadable:
            data["Reference"] = [
                f"Based on GRCh{stats.reference_genome.build}, downloadable"
            ]
        elif stats.reference_genome.status == ReferenceStatus.Unknown:
            data["Reference"] = [
                f"Likely based on GRCh{stats.reference_genome.build}. unknown"
            ]
        data["Gender"] = [stats.gender.name]
        data["Sorted"] = [stats.sorted.name]
        data["Mitochondrial DNA Model"] = [stats.mitochondrial_dna_model.name]

        for key, value in label_map.items():
            if value == "Path":
                continue
            if value not in data or data[value] is None:
                data[value] = [str(stats.__dict__[key])]
        data = TabularData(None, [TabularDataRow(x[0], x[1]) for x in data.items()])
        return data
